(***********************************************************************)
(*                                                                     *)
(*                       Banana Algebra Code                           *)
(*                          (XML edition)                              *)
(*                                                                     *)
(*                         Jacob Andersen                              *)
(*               BRICS, University of Aarhus, Denmark                  *)
(*                                                                     *)
(*     (C)opyright 2007 BRICS / Basic Research in Computer Science     *)
(*                      All rights reserved                            *)
(*                                                                     *)
(***********************************************************************)


open Ast

(* -- library functions ---------------------------------------------- *)

let make = String.make

let iter = List.iter

(* -- general file output -------------------------------------------- *)

let file = ref stdout  

let fprint str : unit = Printf.fprintf !file "%s" str
let fprintln str : unit = Printf.fprintf !file "%s\n" str


(* -- code_xsugar - generate xsugar (.xsg) file from Ast.lcon -------- *)

let rec
xsugar_term (var,regexp,_) =
  begin
    fprint ("R" ^ var) ;
    fprint " = " ;
    xsugar_regexp regexp ;
    fprintln ""
  end

and
escapestring s = 
  let a = Array.make (String.length s) ' ' in
  let i = ref 0 in
  begin
    String.iter (fun c -> begin a.(!i) <- c ; i := (!i) + 1 end) s ;
    Array.fold_right (^) 
                (Array.map (function '\"' -> "\\\""
                                   | '\n' -> "\\n"
                                   | '\r' -> "\\r"
                                   | '\t' -> "\\t"
                                   | c -> String.make 1 c) a)
                ""
  end

and
xsugar_regexp = function
    Union(r1,r2,_)      -> begin
                             fprint "(" ;
                             xsugar_regexp r1 ;
                             fprint "|" ;
                             xsugar_regexp r2 ;
                             fprint ")"
                           end
  | Concat(r1,r2,_)     -> begin
                             fprint "(" ;
                             xsugar_regexp r1 ;
                             xsugar_regexp r2 ;
                             fprint ")"
                           end
  | Optional(r,_)       -> begin
                             xsugar_regexp r ;
                             fprint "?"
                           end
  | Star(r,_)           -> begin
                             xsugar_regexp r ;
                             fprint "*"
                           end
  | Plus(r,_)           -> begin
                             xsugar_regexp r ;
                             fprint "+"
                           end
  | CharClass(ccs,_)    -> begin
                             fprint "[" ;
                             iter xsugar_charclass ccs ;
                             fprint "]"
                           end
  | NegCharClass(ccs,_) ->  begin
                             fprint "[^" ;
                             iter xsugar_charclass ccs ;
                             fprint "]"
                           end
  | Dot(_)              -> fprint "."
  | Empty(_)            -> fprint "#"
  | String(s,_)         -> begin
                             fprint "\"" ;
                             fprint (escapestring s) ;
                             fprint "\""
                           end
  | RegexpID(var,_)     -> begin
                             fprint "<" ;
                             fprint ("R" ^ var) ;
                             fprint ">"
                           end
  | Ch(c,_)             -> xsugar_char c

and
xsugar_charclass = function
    CharExp(c,_)   -> xsugar_char c
  | Range(c1,c2,_) -> begin
                        xsugar_char c1 ;
                        fprint "-" ;
                        xsugar_char c2
                      end

and
xsugar_char c = 
  let s = match c with
      ' '  -> "\\ "
    | '.'  -> "\\."
    | ';'  -> "\\;"
    | '\"' -> "\\\""
    | '#'  -> "\\#"
    | '@'  -> "\\@"
    | '|'  -> "\\|"
    | '&'  -> "\\&"
    | '?'  -> "\\?"
    | '*'  -> "\\*"
    | '+'  -> "\\+"
    | '{'  -> "\\{"
    | '}'  -> "\\}"
    | '('  -> "\\("
    | ')'  -> "\\)"
    | '<'  -> "\\<"
    | '>'  -> "\\>"
    | '['  -> "\\["
    | ']'  -> "\\]"
    | '-'  -> "\\-"
    | '\\' -> "\\\\"
    | '\t' -> "\\t"
    | '\n' -> "\\n"
    | '\r' -> "\\r"
    | c    -> Char.escaped c
  in
    fprint s

and
xsugar_start ps prevN = match ps with
    []  -> ()
  | (n,_,_,_)::ps ->
      begin
        if prevN <> n then
          begin
            if prevN = "" then
              fprint "start... : \"" 
            else
              fprint "         : \""  ;
            fprint (n ^ ":\" [W.] [" ^ ("n" ^ n) ^ " x] [W.] = [") ;
            fprintln (("n" ^ n) ^ " x]") ;
          end ;
        xsugar_start ps n
      end

and
xsugar_productions ps prevN = match ps with
    []  -> ()
  | (n,p,es,_)::ps ->
      begin
        if prevN <> n then
          begin
            fprintln "" ;
            fprint (("n" ^ n) ^ "  : [") ;
            fprint (("n" ^ n) ^ ". x] = <") ;
            fprint (n ^ "> [") ;
            fprintln (("n" ^ n) ^ ". x] </>") ;
            fprint (("n" ^ n) ^ ". : ")
          end
        else
          fprint ((make ((String.length n) + 3) ' ') ^ ": ") ;

        fprint (xsugar_left es 1) ;
        fprint ("= <" ^ n ^ "." ^ p ^ ">") ;

        fprint (xsugar_right es 1) ;
        fprintln " </>" ;

        xsugar_productions ps n
      end

and 
xsugar_left es i = match es with
    [] -> ""
  | e::es -> (if i>1 then "[W.] " else "") ^ (xsugar_left_entity e i)
                          ^ " " ^ (xsugar_left es (i+1))
and
xsugar_left_entity e i =
  let
    escapeampersands s = 
      let a = Array.make (String.length s) ' ' in
      let i = ref 0 in
      begin
        String.iter (fun c -> begin a.(!i) <- c ; i := (!i) + 1 end) s ;
        Array.fold_right (^) 
                 (Array.map (function '&' -> "\\&" | c -> String.make 1 c) a)
                 ""
      end
  in
    match e with
      | Es(s,_)   -> ("\"" ^ (escapeampersands (String.escaped s)) ^ "\"")
      | En(var,_) -> "[" ^ ("n" ^ var) ^ " x" ^ (string_of_int i) ^ "]"

and
xsugar_right es i = match es with
    [] -> ""
  | e::es -> (xsugar_right_entity e i) ^ (xsugar_right es (i+1))

and
xsugar_right_entity e i = match e with
  | Es(s,_)   -> ""
  | En(var,_) -> " [" ^ ("n" ^ var) ^ " x" ^ (string_of_int i) ^ "]"

and
xsugar_regexp_prod (var,_,_) =
  begin
    fprint (("n" ^ var) ^ "  : [") ;
    fprint (("R" ^ var) ^ " x] = <") ;
    fprint (var ^ " value=[") ;
    fprintln (("R" ^ var) ^ " x]/>") ;
  end

and
code_xsugar (((w,wpos),ts,ps):lcon) ns = 
  begin
    (* Namespace *)
    if ns <> "" then
      fprintln ("xmlns = \"http://www.itu.dk/people/brabrand/banana-algebra" ^ ns ^ "\"") ;

    (* Regular expression definitions *)
    iter xsugar_term ts ;

    fprint "W. = " ;
    xsugar_regexp (match w with Empty p -> String("",p) | _ -> w) ;
    fprintln "" ;
    fprintln "" ;


    (* Start symbol *)
    xsugar_start ps "" ;
    fprintln "" ;

    (* Productions *)
    xsugar_productions ps "" ;
    fprintln "" ;

    (* Regular expression productions *)
    iter xsugar_regexp_prod ts
  end 



(* -- code_dtd - generate DTD from Ast.lcon -------------------------- *)

and
dtd_terminal (var,_,_) =
  begin
    fprintln ("<!ELEMENT " ^ var ^ " EMPTY>") ;
    fprintln ("<!ATTLIST " ^ var ^ " value CDATA #REQUIRED>")
  end

and
dtd_production (n,p,es,_) =
  begin
    fprintln ("<!ELEMENT " ^ n ^ "." ^ p 
              ^ (let es' = (List.filter (fun s -> s<>"") 
                             (List.map (fun e -> match e with 
                                          Es(_,_)->""
                                        | En(s,_) -> s) es)) in
                 if (List.length es')>0 then
                   (" (" ^ (String.concat "," es') ^ ")>")
                 else 
                   " EMPTY>"))
  end

and
dtd_nonterminals ps = match ps with
    [] -> ()
  | (n,p,_,_)::ps -> begin
                       fprint ("<!ELEMENT " ^ n ^ " (" ^ n ^ "." ^ p ^ "") ;
                       dtd_nonterminal ps n ;
                       fprintln ")>"
                     end


and
dtd_nonterminal ps prevN = match ps with
    [] -> ()
  | (n,p,_,_)::ps -> begin
                       if n <> prevN then
                         begin
                           fprintln ")>" ;
                           fprint ("<!ELEMENT " ^ n ^ " (" ^ n ^ "." ^ p ^ "")
                         end
                       else
                          fprint ("|" ^ n ^ "." ^ p) ;
                       dtd_nonterminal ps n
                     end

and
code_dtd ((_,ts,ps):lcon) = 
  begin
    iter dtd_terminal ts ;
    iter dtd_production ps ;
    dtd_nonterminals ps
  end 



(* -- code_xslt - generate XSL-T from Ast.xcon ----------------------- *)

and
xslt_typing_templates tau ((_,terms,_) as ls) = match tau with
    [] -> ()
  | (st,tt,_)::tau -> begin
                        if List.mem st (List.map (fun (x,_,_) -> x) terms) then
                          xslt_terminal_template st tt
                        else
                          xslt_nonterminal_template st tt ;
                        xslt_typing_templates tau ls
                      end

and
xslt_terminal_template st tt =
  begin
    fprintln "" ;
    fprintln ("  <xsl:template match=\"s:" ^ st ^ "\">") ;
    fprintln ("    <" ^ tt ^ " value=\"{@value}\" />") ;
    fprintln ("  </xsl:template>")
  end

and
xslt_nonterminal_template st tt =
  begin
    fprintln "" ;
    fprintln ("  <xsl:template match=\"s:" ^ st ^ "\">") ;
    fprintln ("    <xsl:apply-templates />") ;
    fprintln ("  </xsl:template>")
  end

and
xslt_transformer_template (n,p,ast,_) =
  begin
    fprintln "" ;
    fprintln ("  <xsl:template match=\"s:" ^ n ^ "." ^ p ^ "\">") ;
    xslt_ast 4 ast ;
    fprintln ("  </xsl:template>")
  end

and
xslt_ast indent ast = match ast with
    Aast(n,p,asts,_) ->
      begin
        fprintln ((make indent ' ') ^ "<" ^ n ^ ">") ;
        fprintln ((make (indent+2) ' ') ^ "<" ^ n ^ "." ^ p ^ ">") ;
        iter (xslt_ast (indent+4)) asts ;
        fprintln ((make (indent+2) ' ') ^ "</" ^ n ^ "." ^ p ^ ">") ;
        fprintln ((make indent ' ') ^ "</" ^ n ^ ">")
      end
  | Aarg(i,_) ->
      fprintln ((make indent ' ')
                ^ "<xsl:apply-templates select=\"node()[position()="
                ^ (string_of_int i) ^ "]\" />")
  | Aterm(r,s,_) ->
      fprintln ((make indent ' ') ^ "<" ^ r ^ " value=\""
                ^ s ^ "\" />")
  | Aliteral(s,_) -> raise (Failure ("Literal found in AST: \"" ^ s ^ "\""))

and
code_xslt ((xprods,tau,ls,_):xcon) = 
  begin
    fprintln "<xsl:stylesheet" ;
    fprintln "    xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\"" ;
    fprintln "    version=\"1.0\"" ;
    fprintln "    xmlns:s=\"http://www.itu.dk/people/brabrand/banana-algebra/source\"" ;
    fprintln "    xmlns=\"http://www.itu.dk/people/brabrand/banana-algebra/target\">" ;
    fprintln ("  <xsl:output method=\"xml\" doctype-system=\"target.dtd\" " 
              ^ "indent=\"yes\" />") ;
    xslt_typing_templates tau ls ;
    iter xslt_transformer_template xprods ;
    fprintln "</xsl:stylesheet>"    
  end 

(* -- code ----------------------------------------------------------- *)

and
code (((xprods,tau,ls,lt):xcon) as xcon) filename = 
  begin

    (* source.xsg *)
    file := open_out ("source.xsg") ;
    code_xsugar ls "source" ;
    close_out !file ; 

    (* target.xsg *)
    file := open_out ("target.xsg") ;
    code_xsugar lt "target" ;
    close_out !file ; 

    (* source.dtd *)
    file := open_out ("source.dtd") ;
    code_dtd ls ;
    close_out !file ; 

    (* target.dtd *)
    file := open_out ("target.dtd") ;
    code_dtd lt ;
    close_out !file ; 

    (* transform.xsl *)
    file := open_out ("transform.xslt") ;
    code_xslt xcon ;
    close_out !file 
  end

and
writeXSG f lcon ns =
  begin
    file := f ;
    code_xsugar lcon ns
  end
