/***********************************************************************/
/*                                                                     */
/*                       Banana Algebra Parser                         */
/*                                                                     */
/*                 Claus Brabrand, Jacob Andersen                      */
/*               BRICS, University of Aarhus, Denmark                  */
/*                                                                     */
/*     (C)opyright 2007 BRICS / Basic Research in Computer Science     */
/*                      All rights reserved                            */
/*                                                                     */
/***********************************************************************/


%{

  open Ast

  let pos() = (Parsing.symbol_start_pos(), Parsing.symbol_end_pos())

  let assemble f l = 
    List.fold_right (fun x -> (function None   -> (fun l -> l)
                                      | Some x -> (fun l -> x::l))(f x))
                    l []

  (* Default whitespace *)

  let defaultws = Empty((Lexing.dummy_pos,Lexing.dummy_pos))

%}


/* -- Lexeme declarations -------------------------------------------- */

%token LET LETX IN SRC TGT IDX COMMA EQUAL ARROW PLUS BAR BACKSLASH
%token DOLLAR DOT COLON SEMIC HASH MINUS STAR QUESTION
%token LPAR RPAR LBRACE RBRACE LBRACK LBRACKHAT RBRACK LT GT LBANANA RBANANA

%token EOF
%token <int>    ARG
%token <string> STR MLSTR ID
%token <char>   CHAR


/* -- Assoc & Prec --------------------------------------------------- */

%left IN
%left BAR
%left PLUS

/* -- Entry ---------------------------------------------------------- */

%start xexp
%start lexp
%type <Ast.xexp> xexp
%type <Ast.lexp> lexp

%%


/* -- L constants ---------------------------------------------------- */

lcon : LBRACE w lprods RBRACE        { ($2,(assemble fst $3),
                                        (assemble snd $3)) }

lprods : lprod lprods                { $1 :: $2 }
       |                             { [] }

lprod : ID DOT ID COLON es SEMIC     { (None, Some ($1,$3,$5,pos())) }
      | ID EQUAL regexp SEMIC        { (Some ($1,$3,pos()), None) }

es : e es                            { $1 :: $2 }
   |                                 { [] }

e : STR                              { Es($1,pos()) }
  | ID                               { En($1,pos()) }

w : DOLLAR EQUAL regexp SEMIC        { ($3,pos()) }
  |                                  { (defaultws,pos()) }

/* -- regexp --------------------------------------------------------- */

regexp : unionexp                    { $1 }

unionexp : concatexp BAR unionexp    { Union($1,$3,pos()) }
         | concatexp                 { $1 }

concatexp : repeatexp concatexp      { Concat($1,$2,pos()) }
          | repeatexp                { $1 }

repeatexp : repeatexp QUESTION                     { Optional($1,pos()) }
          | repeatexp STAR                         { Star($1,pos()) }
          | repeatexp PLUS                         { Plus($1,pos()) }
          | charclassexp                           { $1 }

charclassexp : LBRACK charclasses RBRACK      { CharClass($2,pos()) }
             | LBRACKHAT charclasses RBRACK   { NegCharClass($2,pos()) }
             | simpleexp                      { $1 }

charclasses : charclass charclasses  { $1 :: $2 } 
            | charclass              { [$1] }

charclass : CHAR                     { CharExp($1,pos()) }
          | CHAR MINUS CHAR          { Range($1,$3,pos()) }

simpleexp : CHAR                     { Ch($1,pos()) }
          | DOT                      { Dot(pos()) }
          | HASH                     { Empty(pos()) }
          | STR                      { String($1,pos()) }
          | LT ID GT                 { RegexpID($2,pos()) }
          | LPAR RPAR                { String("",pos()) }
          | LPAR unionexp RPAR       { $2 }


/* -- Chi constants -------------------------------------------------- */

xprods : xprod xprods                { $1 :: $2 }
       |                             { [] }

xprod : ID DOT ID EQUAL ast SEMIC    { ($1,$3,$5,pos()) }

ast : ID DOT ID LPAR asts RPAR       { Aast($1,$3,$5,pos()) }
    | ARG                            { Aarg($1,pos()) }
    | ID LPAR STR RPAR               { Aterm($1,$3,pos()) }
    | MLSTR                          { Aliteral($1,pos()) }

asts : asts1                         { $1 }
     |                               { [] }

asts1 : ast COMMA asts1              { $1 :: $3 }
      | ast                          { [$1] }


/* -- tau -------------------------------------------------- */

tau : LBRACK maps RBRACK             { $2 }

maps : maps1                         { $1 }
     |                               { [] }

maps1 : map COMMA maps1              { $1 :: $3 }
      | map                          { [$1] }

map : ID ARROW ID                    { ($1,$3,pos()) }


/* -- lexp ----------------------------------------------------------- */

lexp : lcon                          { Lcon($1,pos()) }
     | ID                            { Lvar($1,pos()) }
     | LET ID EQUAL lexp IN lexp     { Llet($2,$4,$6,pos()) }
     | LETX ID EQUAL xexp IN lexp    { Lletx($2,$4,$6,pos()) }
     | lexp PLUS lexp                { Lplus($1,$3,pos()) }
     | lexp BACKSLASH lexp           { Lminus($1,$3,pos()) }
     | SRC LPAR xexp RPAR            { Lsrc($3,pos()) }
     | TGT LPAR xexp RPAR            { Ltgt($3,pos()) }
     | STR                           { Linclude($1,pos()) }
     | LPAR lexp RPAR                { $2 }


/* -- xexp ----------------------------------------------------------- */

xexp : LBANANA lexp ARROW lexp tau xprods RBANANA
                                     { Xcon($6,$5,$2,$4,pos()) }
     | ID                            { Xvar($1,pos()) }
     | LET ID EQUAL lexp IN xexp     { Xlet($2,$4,$6,pos()) }
     | LETX ID EQUAL xexp IN xexp    { Xletx($2,$4,$6,pos()) }
     | xexp PLUS xexp                { Xplus($1,$3,pos()) }
     | xexp BAR xexp                 { Xseq($1,$3,pos()) }
     | IDX LPAR lexp RPAR            { Xidx($3,pos()) }
     | STR                           { Xinclude($1,pos()) }
     | LPAR xexp RPAR                { $2 }
	   
