(***********************************************************************)
(*                                                                     *)
(*                      Banana Algebra Pretty                          *)
(*                                                                     *)
(*                 Claus Brabrand, Jacob Andersen                      *)
(*               BRICS, University of Aarhus, Denmark                  *)
(*                                                                     *)
(*     (C)opyright 2007 BRICS / Basic Research in Computer Science     *)
(*                      All rights reserved                            *)
(*                                                                     *)
(***********************************************************************)


open Ast

let dp = (Lexing.dummy_pos, Lexing.dummy_pos)

(* -- general file output -------------------------------------------- *)

let file = ref stdout  
let i_level = ref 0

let fprint str : unit = Printf.fprintf !file "%s" str
let fnewline () : unit = Printf.fprintf !file "\n%s" (String.make !i_level ' ')

let indent () : unit = i_level := (!i_level + 2)
let outdent () : unit = i_level := (!i_level - 2)

(* -- L constants ---------------------------------------------------- *)


let rec 
p_prod (n,p,es,_) =
  begin
    fnewline () ;
    fprint n ;
    fprint "." ;
    fprint p ;
    fprint " :  " ;
    List.iter p_entity es ;
    fprint ";"
  end

and
p_entity = function
    Es(s,_) -> fprint ("\"" ^ (String.escaped s) ^ "\" ")
  | En(n,_) -> fprint (n ^ " ")

and
p_term (t,r,_) =
  begin
    fnewline () ;
    fprint t ;
    fprint " = " ;
    p_regexp r ;
    fprint " ;" ;
  end

(* -- regexp --------------------------------------------------------- *)

and
p_regexp = function
    Union(r1,r2,_)        -> begin
                               fprint "(" ;
                               p_regexp r1 ;
                               fprint "|" ;
                               p_regexp r2 ;
                               fprint " )"
                             end
  | Concat(r1,r2,_)       -> begin
                               fprint "(" ;
                               p_regexp r1 ;
                               p_regexp r2 ;
                               fprint " )"
                             end
  | Optional(r,_)        ->  begin
                               fprint "(" ;
                               p_regexp r ;
                               fprint ")?"
                             end
  | Star(r,_)            ->  begin
                               fprint "(" ;
                               p_regexp r ;
                               fprint ")*"
                             end
  | Plus(r,_)             -> begin
                               fprint "(" ;
                               p_regexp r ;
                               fprint ")+"
                             end
  | CharClass(l,_)        -> begin
                               fprint "[" ;
                               List.iter p_charclass l ;
                               fprint "]"
                             end
  | NegCharClass(l,_)     -> begin
                               fprint "[^" ;
                               List.iter p_charclass l ;
                               fprint "]"
                             end
  | Dot(_)                -> fprint "."
  | Empty(_)              -> fprint "#"
  | String(s,_)           -> fprint ("\"" ^ (String.escaped s) ^ "\"")
  | RegexpID(v,_)         -> begin
                               fprint "<" ;
                               fprint v ;
                               fprint ">"
                             end
  | Ch(c,_)               -> begin
                               if c='.' || c='#' || c='+' || c='*' ||
                                  c='?' || c='|' || c='(' || c=')' ||
                                  c='[' || c='^' || c='<' || c=';' ||
                                  c=' ' || c='/' || c='\"' then
                                 fprint "\\" ;
                               if c='\r' then
                                 fprint "\\r"
                               else
                                 fprint (Char.escaped c)
                             end

and
p_charclass = function
    Range(c1,c2,_) -> begin
                        p_char c1 ;
                        fprint "-" ;
                        p_char c2
                      end
  | CharExp(c,_)   -> p_char c

and
p_char c = begin
             if c='^' || c='-' || c=' ' || c='/' then
               fprint "\\" ;
             if c='\r' then
               fprint "\\r"
             else
               fprint (Char.escaped c)
           end


(* -- X constants ---------------------------------------------------- *)

and 
p_xprod (n,p,a,_) =
  begin
    fnewline () ;
    fprint n ;
    fprint "." ;
    fprint p ;
    fprint "  = " ;
    p_ast a ;
    fprint "  ;" ;
  end

and
p_ast = function
    Aast(n,p,aa,_) -> begin 
                        fprint n ;
                        fprint "." ;
                        fprint p ;
                        fprint "(" ;
                        p_asts aa ;
                        fprint ")"
                      end
  | Aarg(i,_)      -> begin 
                        fprint "$" ;
                        fprint (string_of_int i) ;
                      end
  | Aterm(t,s,_)   -> begin 
                        fprint t ;
                        fprint "(\"" ;
                        fprint (String.escaped s) ;
                        fprint "\")" 
                      end
  | Aliteral(s,_)  -> fprint ("'" ^ (escapequotes s) ^ "'")


and
escapequotes s = 
  let a = Array.make (String.length s) ' ' in
  let i = ref 0 in
  begin
    String.iter (fun c -> begin a.(!i) <- c ; i := (!i) + 1 end) s ;
    Array.fold_right (^) 
                (Array.map (function '\'' -> "''" | c -> String.make 1 c) a)
                ""
  end

and
p_asts = function
    []   -> ()
  | a::aa -> begin
               p_ast a ;
               if (aa != []) then fprint ", " ;
               p_asts aa
             end

and
p_tau = function
    []        -> ()
  | [(s,t,_)] -> begin
                   fprint s ;
                   fprint " -> " ;
                   fprint t ;
                 end
  | ts        -> let rec
                   aux = function
                       []        -> ()
                     | (s,t,_)::ts -> begin
                                        fnewline () ;
                                        fprint s ;
                                        fprint " -> " ;
                                        fprint t ;
                                        if (ts != []) then fprint " ," ;
                                        aux ts
                                      end
                 in
                   begin
                     indent () ;
                     aux ts ;
                     outdent () ;
                     fnewline () ;
                   end

(* -- lexp ----------------------------------------------------------- *)

and
p_lexp = function
    Lcon(((w,wpos),ts,ps),_) -> begin
                           indent ();
                           fnewline () ;
                           fprint "{" ;
                           indent () ;
                           p_term ("$",w,wpos) ;
                           List.iter p_term ts ;
                           List.iter p_prod ps ;
                           outdent() ;
                           fnewline ();
                           fprint "}" ;
                           outdent () ;
                           fnewline ()
                         end
  | Lvar(v,_) -> fprint v
  | Llet(v,lexp,lexp',_) -> begin
                               indent ();
                               fnewline () ;
                               fprint "let" ;
                               indent () ;
                               fnewline () ;
                               fprint v ;
                               fprint " = " ;
                               indent () ;
                               p_lexp lexp ;
                               outdent () ;
                               outdent () ;
                               fnewline () ;
                               fprint "in" ;
                               indent () ;
                               fnewline () ;
                               fprint "(" ;
                               p_lexp lexp' ;
                               fprint ")" ;
                               outdent () ;
                               outdent () ;
                               fnewline ()
                             end
  | Lletx(v,xexp,lexp',_) -> begin
                               indent ();
                               fnewline () ;
                               fprint "letx" ;
                               indent () ;
                               fnewline () ;
                               fprint v ;
                               fprint " = " ;
                               indent () ;
                               p_xexp xexp ;
                               outdent () ;
                               outdent () ;
                               fnewline () ;
                               fprint "in" ;
                               indent () ;
                               fnewline () ;
                               fprint "(" ;
                               p_lexp lexp' ;
                               fprint ")" ;
                               outdent () ;
                               outdent () ;
                               fnewline ()
                             end
  | Lplus(lexp,lexp',_) -> begin
                             fprint "(" ;
                             p_lexp lexp ;
                             fprint " + " ;
                             p_lexp lexp' ;
                             fprint ")" ;
                           end
  | Lminus(lexp,lexp',_) -> begin
                              fprint "(" ;
                              p_lexp lexp ;
                              fprint " \\ " ;
                              p_lexp lexp' ;
                              fprint ")" ;
                            end
  | Lsrc(xexp,_) -> begin
                      fprint "src(" ;
                      p_xexp xexp ;
                      fprint ")" 
                    end
  | Ltgt(xexp,_) -> begin
                      fprint "tgt(" ;
                      p_xexp xexp ;
                      fprint ")" 
                    end
  | Linclude(s,_) ->  fprint ("\"" ^ (String.escaped s) ^ "\"")


(* -- xexp ----------------------------------------------------------- *)

and
p_xexp = function
    Xcon(xs,t,ls,lt,_) -> begin
                            indent ();
                            fnewline () ;
                            fprint "(|" ;
                            p_lexp ls ;
                            fprint " -> " ;
                            p_lexp lt ;
                            indent () ;
                            fnewline () ;
                            fprint "[" ;
                            p_tau t ;
                            fprint "]" ;
                            fnewline () ;
                            List.iter p_xprod xs ;
                            outdent() ;
                            fnewline ();
                            fprint "|)" ;
                            outdent () ;
                            fnewline ()
                          end
  | Xvar(v,_) -> fprint v
  | Xlet(v,lexp,xexp',_) -> begin
                              indent () ;
                              fnewline () ;
                              fprint "let" ;
                              indent () ;
                              fnewline () ;
                              fprint v ;
                              fprint " = " ;
                              indent () ;
                              p_lexp lexp ;
                              outdent () ;
                              outdent () ;
                              fnewline () ;
                              fprint "in" ;
                              indent () ;
                              fnewline () ;
                              fprint "(" ;
                              p_xexp xexp' ;
                              fprint ")" ;
                              outdent () ;
                              outdent () ;
                              fnewline ()
                            end
  | Xletx(v,xexp,xexp',_) -> begin
                               indent ();
                               fnewline () ;
                               fprint "letx" ;
                               indent () ;
                               fnewline () ;
                               fprint v ;
                               fprint " = " ;
                               indent () ;
                               p_xexp xexp ;
                               outdent () ;
                               outdent () ;
                               fnewline () ;
                               fprint "in" ;
                               indent () ;
                               fnewline () ;
                               fprint "(" ;
                               p_xexp xexp' ;
                               fprint ")" ;
                               outdent () ;
                               outdent () ;
                               fnewline ()
                             end
  | Xplus(xexp,xexp',_) -> begin
                             fprint "(" ;
                             p_xexp xexp ;
                             fprint " + " ;
                             p_xexp xexp' ;
                             fprint ")" ;
                           end
  | Xseq(xexp,xexp',_) -> begin
                            fprint "(" ;
                            p_xexp xexp ;
                            fprint " | " ;
                            p_xexp xexp' ;
                            fprint ")" ;
                          end
  | Xidx(lexp,_) -> begin
                      fprint "idx(" ;
                      p_lexp lexp ;
                      fprint ")" 
                    end
  | Xinclude(s,_) -> fprint ("\"" ^ (String.escaped s) ^ "\"")


(* -- entry points --------------------------------------------------- *)

and
pretty_xcon ((xprods,tau,ls,lt):xcon) (f:out_channel) :unit =
  begin
    i_level := 0 ;
    file := f ;
    p_xexp (Xcon(xprods, tau, Lcon(ls,dp), Lcon(lt,dp), dp))
  end

and 
pretty_lcon (lcon:lcon) (f:out_channel) :unit =
  begin
    i_level := 0 ;
    file := f ;
    p_lexp (Lcon(lcon,dp))
  end

and
pretty_xexp (xexp:xexp) (f:out_channel) :unit =
  begin
    i_level := 0 ;
    file := f ;
    p_xexp xexp
  end

and
pretty_lexp (lexp:lexp) (f:out_channel) :unit =
  begin
    i_level := 0 ;
    file := f ;
    p_lexp lexp
  end
