(***********************************************************************)
(*                                                                     *)
(*                     Banana Algebra Transform                        *)
(*                                                                     *)
(*                 Claus Brabrand, Jacob Andersen                      *)
(*               BRICS, University of Aarhus, Denmark                  *)
(*                                                                     *)
(*     (C)opyright 2007 BRICS / Basic Research in Computer Science     *)
(*                      All rights reserved                            *)
(*                                                                     *)
(***********************************************************************)

open Ast

open List

(* -- Misc ----------------------------------------------------------- *)

let rec
get_tau (tau : tau) (n:string) : string = match tau with
    [] -> raise Not_found
  | (a,b,_)::tau -> if (a = n)
                    then b
                    else get_tau tau n

and
get_language ((_,_,ps):lcon) (n:var) (p:var) : prod =
  try
    find (fun (n1,p1,_,_) -> n=n1 && p=p1) ps
  with
    Not_found ->
      raise (Invalid_argument("'" ^ n ^ "." ^ p ^ "'"))


(* -- check ast ------------------------------------------------------ *)

let rec
check_ast (tenv :string list) (l:lcon) (nt:string) (ast:ast) = match ast with
    Aast(n,p,asts,pos) -> if n <> nt then
                            raise (Error("Wrong type. Expected '" ^ nt ^
                                         "' - got '" ^ n ^ "'",pos))
                          else
                            let
                              (_,_,es,_) =
                                try
                                  get_language l n p
                                with Invalid_argument s ->
                                  raise (Error ("Unknown production: "^s,pos))
                            in
                            check_asts tenv l es asts pos
  | Aarg(i,pos)        -> begin
                            try
                              if ((nth tenv (i-1)) <> nt) then
                                let n = nth tenv (i-1) in
                                if n.[0] = '*' then
                                  let (posl : pos) = Marshal.from_string n
                                          ((String.index_from n 1 '*') + 1) in
                                  let n = String.sub n 1
                                          ((String.index_from n 1 '*') - 1) in
                                  raise (Conflict("Source language " ^
                                    "nonterminal '" ^ n ^ "' not found in " ^
                                    "transformation typing - Expected type '" ^
                                    nt ^ "'",
                                  pos,posl))

                                else
                                  raise (Error("Wrong type. Expected '" ^ nt ^
                                         "' - got '" ^ n ^ "'", pos))
                            with Failure _ ->
                              raise (Error("Argument index '$" ^
                                    (string_of_int i) ^ "' out of bounds",pos))
                          end
  | Aterm(t,s,pos)     -> if t <> nt then
                            raise (Error("Wrong type. Expected '" ^ nt ^
                                         "' - got '" ^ t ^ "'",pos))
                     (* TODO: check that s conforms to regexp for t *)
  | Aliteral _         -> raise (Failure ("Literal Chi expression should " ^
                                 "have been parsed during normalization"))

and
check_asts (tenv :string list) (l:lcon) (es:entity list)
           (asts:ast list) (pos:pos) = match es with
    []             -> if (asts <> []) then
                        raise (Error("Too many arguments",pos))
  | Es(s,_)::es    -> check_asts tenv l es asts pos
  | En(n,pos')::es -> begin
                        if (asts = []) then
                          raise (Error("Too few arguments",pos)) ;
                        check_ast tenv l n (hd asts) ;
                        check_asts tenv l es (tl asts) pos
                      end

(* -- transform ------------------------------------------------------ *)

(* TODO: think this through!!! *)
let rec
transform ((chi,tau,ls,lt):xcon) (ast:ast) : ast = match ast with
    Aast(n,p,asts,pos) -> let
                            gamma = map (transform (chi,tau,ls,lt)) asts
                          in
                            let
                              (_,_,a,_) =
                                try
                                  (find (fun (n',p',_,_)-> n=n' && p=p') chi)
                                with Not_found ->
                                  raise (Error ("'" ^ n ^ "." ^ p ^ "' " ^ 
                                    "undefined in subsequent transformation",
                                    pos))
                            in
                              copy_substitute gamma a
  | Aarg(i,pos)      -> Aarg(i,pos) (* leave arguments *)
  | Aterm(t,s,pos)   -> Aterm(get_tau tau t,s,pos)
  | Aliteral _       -> raise (Failure ("Literal Chi expression should have " ^
                                        "been parsed during normalization"))

and
copy_substitute (gamma:ast list) (ast:ast) : ast = match ast with
    Aast(n,p,asts,pos) -> let xasts = map (copy_substitute gamma) asts
                          in Aast(n,p,xasts,pos)
  | Aarg(i,_) -> nth gamma (i-1)
  | Aterm(t,s,pos) -> Aterm(t,s,pos)
  | Aliteral _       -> raise (Failure ("Literal Chi expression should have " ^
                                        "been parsed during normalization"))


(* -- unparse -------------------------------------------------- *)
(*
and
unparse ls ast = match ast with
  Aast(n,p,asts,_) -> let es = get_language ls n p
                      in unparses ls es asts
| Aarg(k,inx,pos)  -> begin
                        print "$" ;
                        print (int_to_string k) ;
                        print ":" ;
                        print (int_to_string inx)
                      end
| Aterm(t,s,pos)   -> begin
                        print "$" ;
                        print t ;
                        print "(\"" ;
                        print s ;
                        print "\")"
                      end

and
unparses ls (es : entity list) (asts : ast list) = match es with
    [] -> ()
  | Es(s,_)::es -> begin
                     print s ;
                     unparses ls es asts
                   end
  | Et(t,_)::es -> begin
                     unparse ls (hd asts) ;
                     unparses ls es (tl asts)
                   end
  | En(n,_)::es -> begin
                     unparse ls (hd asts) ;
                     unparses ls es (tl asts)
                   end

                   

*)
